/***********************************************************************************************************************
 *
 * blueBill Resources - open source birding
 * Copyright (C) 2009-2011 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://bluebill.tidalwave.it
 * SCM: https://java.net/hg/bluebill~resources-src
 *
 **********************************************************************************************************************/
package it.tidalwave.bluebill.factsheet.wikipedia.photos;

import java.io.IOException;
import javax.annotation.Nonnull;
import it.tidalwave.bluebill.taxonomy.Taxon;
import it.tidalwave.bluebill.taxonomy.Taxon.Builder;
import it.tidalwave.bluebill.taxonomy.birds.col.CatalogueOfLifeImporter;
import it.tidalwave.openrdf.elmo.impl.ElmoEntityFactory;
import it.tidalwave.semantic.EntityFactory;
import it.tidalwave.semantic.Wrapper;
import it.tidalwave.util.Initializer;
import it.tidalwave.xml.XPathProvider;
import it.tidalwave.xml.XmlParser;
import it.tidalwave.xml.dom4j.Dom4jXmlParser;
import it.tidalwave.xml.jaxen.JaxenXPathProvider;
import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.net.Proxy;
import java.net.URI;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import javax.xml.namespace.QName;
import lombok.Cleanup;
import org.apache.commons.io.IOUtils;
import org.openrdf.concepts.dc.DcResource;
import org.openrdf.concepts.foaf.FoafResource;
import org.openrdf.concepts.foaf.Image;
import org.openrdf.concepts.skos.core.Concept;
import org.xml.sax.EntityResolver;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.jaxen.XPath;
import lombok.extern.slf4j.Slf4j;

/***********************************************************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
@Slf4j
public class WikipediaPhotosImporter extends CatalogueOfLifeImporter
  {
    public static final String DCMI_TYPE_STILL_IMAGE = "http://purl.org/dc/dcmitype/StillImage";
    public static final String LICENSE_GNU_FDL = "http://www.gnu.org/copyleft/fdl.html";
    private static final String URI_COMMONS_WIKIMEDIA = "http://commons.wikimedia.org";
    private static final String URN_BLUEBILL_WIKIMEDIA_CONTRIBUTOR = "urn:bluebill:wikimedia/contributor/";
    
    @Nonnull
    private final transient XmlParser xmlParser;

    @Nonnull
    private final transient XPathProvider xPathProvider;

    private transient Proxy proxy = Proxy.NO_PROXY;

    private static final String XPATH_LI_GALLERY_BOX = "//html:li[@class='gallerybox']";
    private static final String XPATH_A_HREF = ".//html:a[@class='image']/@href";
    private static final String XPATH_A_IMG_SRC = ".//html:a[@class='image']/html:img/@src";
    
    private static final String XPATH_IMAGE_URL = "//html:div[@id='file']/html:a[1]/@href";
    private static final String XPATH_LICENSE_URL = "/html:html/html:head/html:link[@rel='copyright']/@href";
    private static final String XPATH_AUTHOR_P = "//html:tr[html:th='Author']/html:td/html:p/text()";
    
    private static final String XPATH_AUTHOR_A_NAME = "//html:tr[html:th='Author']/html:td/html:p/html:a/text()";
    private static final String XPATH_AUTHOR_A_URI = "//html:tr[html:th='Author']/html:td/html:p/html:a/@href";
    
    private static final String XPATH_AUTHOR_A_NAME3 = "//html:p[contains(html:b, 'Photographer:')]/html:a/text()";
    private static final String XPATH_AUTHOR_A_URI3 = "//html:p[contains(html:b, 'Photographer:')]/html:a/@href";

    private static final String XPATH_AUTHOR_A_NAME5 = "//html:p[contains(., 'Author:')]/html:a/text()";
    private static final String XPATH_AUTHOR_A_URI5 = "//html:p[contains(., 'Author:')]/html:a/@href";

    private static final String XPATH_AUTHOR_A_NAME6 = "//html:li[contains(html:b, 'Fotograf')]/html:a/text()";
    private static final String XPATH_AUTHOR_A_URI6 = "//html:li[contains(html:b, 'Fotograf')]/html:a/@href";

    private static final String XPATH_AUTHOR_A_NAME7 = "//html:i[contains(., 'grants anyone')]/html:a/text()";
    private static final String XPATH_AUTHOR_A_URI7 = "//html:i[contains(., 'grants anyone')]/html:a/@href";

    private static final String XPATH_AUTHOR_A_URI2 = "//html:p[contains(., 'Author:')]/text()";
    
    private static final String XPATH_AUTHOR_A_URI4 = "//html:li[contains(., 'Creator:')]/text()";
    
    protected final transient EntityFactory<Description, ElmoDescription> stillImageFactory =
            new ElmoEntityFactory<Description, ElmoDescription>(Description.class,
                                                    ElmoDescription.class,
                                                    Image.class,
                                                    DcResource.class);

    protected final transient EntityFactory<Organization, ElmoOrganization> organizationFactory =
            new ElmoEntityFactory<Organization, ElmoOrganization>(Organization.class,
                                                              ElmoOrganization.class,
                                                              Concept.class,
                                                              DcResource.class,
                                                              org.openrdf.concepts.foaf.Organization.class,
                                                              FoafResource.class);

    protected final transient EntityFactory<Person, ElmoPerson> authorFactory =
            new ElmoEntityFactory<Person, ElmoPerson>(Person.class,
                                                              ElmoPerson.class,
                                                              Concept.class,
                                                              DcResource.class,
                                                              org.openrdf.concepts.foaf.Person.class);

    protected transient Organization wikiMediaCommons;

    private transient final Map<String, Person> map = new HashMap<String, Person>();

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    protected final transient EntityResolver entityResolver = new EntityResolver()
      {
        @Override @Nonnull
        public InputSource resolveEntity (final @Nonnull String publicId, final @Nonnull String systemId)
          throws SAXException, IOException
          {
            if ("http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd".equals(systemId))
              {
                return new InputSource(getClass().getResourceAsStream("xhtml1-transitional.dtd"));
              }

            return null;
          }
      };

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public WikipediaPhotosImporter()
      throws IOException
      {
        this(new Dom4jXmlParser(), new JaxenXPathProvider());
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    /* package */ WikipediaPhotosImporter (final @Nonnull XmlParser xmlParser,
                                           final @Nonnull XPathProvider xPathProvider)
      throws IOException
      {
        this.xmlParser = xmlParser;
        this.xPathProvider = xPathProvider;
        this.xmlParser.setEntityResolver(entityResolver);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Override
    protected void initialize()
      throws IOException
      {
        super.initialize();
        wikiMediaCommons = organizationFactory.create(new QName("http://commons.wikimedia.org/#organization"));
        wikiMediaCommons.getLookup().lookup(Concept.class).setRdfsLabel("Wikimedia Commons");
//        wikiMediaCommons.as(Concept.class).setRdfsLabel("Wikimedia Commons"); // FIXME: ClassCastException
//        wikiMediaCommons.as(Organization.class).getFoafHomepages().add("http://wikipedia.org");
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    protected Builder process (final @Nonnull String path,
                               final @Nonnull String scientificName,
                               @Nonnull Builder builder)
      throws Exception
      {
        log.info("process({}, {}, {})", new Object[] { path, scientificName, builder });

        if (builder.getRank() != Taxon.Rank.SPECIES)
          {
            log.info(">>>> ignoring rank {}", builder.getRank());
            throw new IgnoreException();
          }

        final URL url = new URL("http://commons.wikimedia.org/wiki/" + scientificName.replace(' ', '_'));
        log.debug(">>>> scraping page: {}", url);
        @Cleanup final InputStream is = url.openStream();
        final List<URI> imagePageUrls = scrapeMainPage(is);
        is.close();
        log.debug(">>>> image page Urls: {}", imagePageUrls);

        builder = builder.withInitializer(new Initializer<Taxon>()
          {
            @Override @Nonnull
            public Taxon initialize (final @Nonnull Taxon taxon)
              {
                for (final URI imagePageUrl : imagePageUrls)
                  {
                    try
                      {
                        log.info(">>>> processing image page: {}", imagePageUrl);
                        final String path = imagePageUrl.getPath().toLowerCase();
                        
                        if (!path.endsWith(".jpg") && !path.endsWith(".jpeg"))
                          {
                            log.warn("Ignoring non-jpeg resource: {}", imagePageUrl);    
                          }
                        else
                          {
                            final InputStream is2 = imagePageUrl.toURL().openStream();
                            final Description media = scrapeImagePage(is2);
                            is2.close();
                            media.getLookup().lookup(Image.class).getFoafDepicts().add(Wrapper.unwrap(taxon));
                          }
                      }
                    catch (Exception e)
                      {
                        log.error("", e);
                      }
                  }
                
                return taxon;
              }
          });

        return builder;
      }

    /*******************************************************************************************************************
     *
     * Scrapes the main page to extract the links of the image pages.
     *
     ******************************************************************************************************************/
    protected List<URI> scrapeMainPage (final @Nonnull InputStream is)
      throws Exception
      {
        final Object document = xmlParser.parse(is);
        final XPath exprLi      = xPathProvider.createXPath(document, XPATH_LI_GALLERY_BOX);
        final XPath exprAHref   = xPathProvider.createXPath(document, XPATH_A_HREF);
        final XPath exprAImgSrc = xPathProvider.createXPath(document, XPATH_A_IMG_SRC);

        final List<Object> liElements = exprLi.selectNodes(document);
        final List<URI> result = new ArrayList<URI>();

        for (int i = 0; i < liElements.size(); i++)
          {
            final Object liElement = liElements.get(i);
            final String wikiPage = URI_COMMONS_WIKIMEDIA + exprAHref.stringValueOf(liElement);
            final String thumbnail = exprAImgSrc.stringValueOf(liElement);
            log.debug("wikiPage: {} thumbnail: {}", wikiPage , thumbnail);

            result.add(new URI(wikiPage));
          }

        return result;
      }

    /*******************************************************************************************************************
     *
     * Scrapes the image page to extract the image link and metadata.
     *
     ******************************************************************************************************************/
    protected Description scrapeImagePage (final @Nonnull InputStream is)
      throws Exception
      {
        final byte[] bytes = IOUtils.toByteArray(is);
//        final String string = new String(bytes, "UTF-8");
        final Object document = xmlParser.parse(new ByteArrayInputStream(bytes));
        final XPath exprImageUrl   = xPathProvider.createXPath(document, XPATH_IMAGE_URL);
        final XPath exprLicenseUrl = xPathProvider.createXPath(document, XPATH_LICENSE_URL);
        final XPath exprAuthorP = xPathProvider.createXPath(document, XPATH_AUTHOR_P);
        final XPath exprAuthorAName = xPathProvider.createXPath(document, XPATH_AUTHOR_A_NAME);
        final XPath exprAuthorAUri = xPathProvider.createXPath(document, XPATH_AUTHOR_A_URI);
        final XPath exprAuthorAName3 = xPathProvider.createXPath(document, XPATH_AUTHOR_A_NAME3);
        final XPath exprAuthorAUri3 = xPathProvider.createXPath(document, XPATH_AUTHOR_A_URI3);
        final XPath exprAuthorAName5 = xPathProvider.createXPath(document, XPATH_AUTHOR_A_NAME5);
        final XPath exprAuthorAUri5 = xPathProvider.createXPath(document, XPATH_AUTHOR_A_URI5);
        final XPath exprAuthorAName6 = xPathProvider.createXPath(document, XPATH_AUTHOR_A_NAME6);
        final XPath exprAuthorAUri6 = xPathProvider.createXPath(document, XPATH_AUTHOR_A_URI6);
        final XPath exprAuthorAName7 = xPathProvider.createXPath(document, XPATH_AUTHOR_A_NAME7);
        final XPath exprAuthorAUri7 = xPathProvider.createXPath(document, XPATH_AUTHOR_A_URI7);
        final XPath exprAuthorAUri2 = xPathProvider.createXPath(document, XPATH_AUTHOR_A_URI2);
        final XPath exprAuthorAUri4 = xPathProvider.createXPath(document, XPATH_AUTHOR_A_URI4);

        final String imageUrl = exprImageUrl.stringValueOf(document);
        final String licenseUrl = exprLicenseUrl.stringValueOf(document);
//        log.info(">>>> imageUrl: {}, licenseUrl: {}", imageUrl, licenseUrl);
        
        final Description media = stillImageFactory.create(new QName(imageUrl));
        final DcResource dcResource = media.getLookup().lookup(DcResource.class); // FIXME: as()
        dcResource.getDcTypes().add(DCMI_TYPE_STILL_IMAGE);
        dcResource.getDcFormats().add("image/jpeg");
        dcResource.getDcPublishers().add(Wrapper.unwrap(wikiMediaCommons));
        dcResource.getDcIdentifiers().add(imageUrl);
        dcResource.getDcRights().add(licenseUrl);

        final String authorP = exprAuthorP.stringValueOf(document).replaceAll(", *$", "");
        final String authorAName = exprAuthorAName.stringValueOf(document);
        final String authorAUri = exprAuthorAUri.stringValueOf(document);
        final String authorAName3 = exprAuthorAName3.stringValueOf(document);
        final String authorAUri3 = exprAuthorAUri3.stringValueOf(document);
        final String authorAName5 = exprAuthorAName5.stringValueOf(document);
        final String authorAUri5 = exprAuthorAUri5.stringValueOf(document);
        final String authorAName6 = exprAuthorAName6.stringValueOf(document);
        final String authorAUri6 = exprAuthorAUri6.stringValueOf(document);
        final String authorAName7 = exprAuthorAName7.stringValueOf(document);
        final String authorAUri7 = exprAuthorAUri7.stringValueOf(document);
        final String authorAUri2 = exprAuthorAUri2.stringValueOf(document);
        final String authorAUri4 = exprAuthorAUri4.stringValueOf(document);
        log.info(">>>> authorP: {}, authorAName: {}, authorAUri: {}", new Object[] { authorP, authorAName, authorAUri });
        
        if (!"".equals(authorAName))
          {
            dcResource.getDcCreators().add(Wrapper.unwrap(findOrCreateAuthor(fixedUri(URI_COMMONS_WIKIMEDIA, authorAUri), authorAName)));
          }
        else if (!"".equals(authorAName3))
          {
            dcResource.getDcCreators().add(Wrapper.unwrap(findOrCreateAuthor(fixedUri(URI_COMMONS_WIKIMEDIA, authorAUri3), authorAName3)));
          }
        else if (!"".equals(authorAName5))
          {
            dcResource.getDcCreators().add(Wrapper.unwrap(findOrCreateAuthor(fixedUri(URI_COMMONS_WIKIMEDIA, authorAUri5), authorAName5)));
          }
        else if (!"".equals(authorAName6))
          {
            dcResource.getDcCreators().add(Wrapper.unwrap(findOrCreateAuthor(fixedUri(URI_COMMONS_WIKIMEDIA, authorAUri6), authorAName6)));
          }
        else if (!"".equals(authorAName7))
          {
            dcResource.getDcCreators().add(Wrapper.unwrap(findOrCreateAuthor(fixedUri(URI_COMMONS_WIKIMEDIA, authorAUri7), authorAName7)));
          }
        else if (!"".equals(authorP))
          {
            dcResource.getDcCreators().add(Wrapper.unwrap(findOrCreateAuthor(fixedUri(URN_BLUEBILL_WIKIMEDIA_CONTRIBUTOR, authorP), authorP)));
          }
        else if (!"".equals(authorAUri2) && !"Author:".equals(authorAUri2.trim()))
          {
            final String displayName = authorAUri2.replace("Author:", "").trim();
            final String uri = URN_BLUEBILL_WIKIMEDIA_CONTRIBUTOR + authorAUri2.replace("Author:", "").replaceAll("\\(.*\\)", "").replace(" ", "").trim();
            dcResource.getDcCreators().add(Wrapper.unwrap(findOrCreateAuthor(uri, displayName)));
          }
        else if (!"".equals(authorAUri4))
          {
            final String displayName = authorAUri4.replace("Creator:", "").trim();
            final String uri = URN_BLUEBILL_WIKIMEDIA_CONTRIBUTOR + authorAUri4.replace("Creator:", "").replaceAll("\\(.*\\)", "").replace(" ", "").trim();
            dcResource.getDcCreators().add(Wrapper.unwrap(findOrCreateAuthor(uri, displayName)));
          }
        else
          {
            log.warn("No author for {}", imageUrl);  
          }
        
//        int h = string.indexOf("Creator:");
//        
//        if (h >= 0)
//          {
//            h += "Creator:".length();
//            int j = string.substring(h).indexOf("</");
//            String author = string.substring(h, h + j).trim();
//            log.info("AUTHOR: {}", author);
//            final String authorUri = "http://wikipedia.org/contributor/" + author;
//            dcResource.getDcCreators().add(Wrapper.unwrap(findOrCreateAuthor(authorUri, author)));
//          }
//        
//        else
//          {
//            int i = string.indexOf("Author:");
//
//            if (i >= 0)
//              {
//                i += "Author:".length();
//                int j = string.substring(i).indexOf("</p>");
//                String author = string.substring(i, i + j).trim();
//
//                if (author.contains("href"))
//                  {
//                    author = author.replace("</b> ", "");
//                    int i1 = author.indexOf('"');
//                    int i2 = author.substring(i1 + 1).indexOf('"') + i1;
//                    final String authorUri = "http://commons.wikimedia.org" + author.substring(i1 + 1, i2 + 1);
//                    int j1 = author.indexOf('>');
//                    int j2 = author.substring(j1 + 1).indexOf('<') + j1;
//                    final String authorName = author.substring(j1 + 1, j2 + 1);
//                    dcResource.getDcCreators().add(Wrapper.unwrap(findOrCreateAuthor(authorUri, authorName)));
//                    log.info("AUTHOR: {}", authorName);
//                  }
//                else
//                  {
//                    log.info("AUTHOR: {}", author);
//                    final String authorUri = "http://wikipedia.org/contributor/" + author;
//                    dcResource.getDcCreators().add(Wrapper.unwrap(findOrCreateAuthor(authorUri, author)));
//                  }
//              }
//            
//            else
//              {
//                int j = string.indexOf("Author</th>");
//
//                if (j >= 0)
//                  {
//                    final String author = string.substring(j + 12);
//                    int i1 = author.indexOf("href=\"") + 5;
//                    int i2 = author.substring(i1 + 1).indexOf('"') + i1;
//                    final String authorUri = "http://commons.wikimedia.org" + author.substring(i1 + 1, i2 + 1);
//                    int j1 = author.indexOf("\">") + 1;
//                    int j2 = author.substring(j1 + 1).indexOf('<') + j1;
//                    final String authorName = author.substring(j1 + 1, j2 + 1);
//                    dcResource.getDcCreators().add(Wrapper.unwrap(findOrCreateAuthor(authorUri, authorName)));
//                    log.info("AUTHOR: {}", authorName);
//                  }
//              }
//          }

        return media;
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    private static String fixedUri (final @Nonnull String prefix, final @Nonnull String s)
      {
        return s.startsWith("http://") ? s : prefix + s;
      }
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    private Person findOrCreateAuthor (final @Nonnull String uri, final @Nonnull String name)
      {
        Person foafEntity = map.get(uri);

        if (foafEntity == null)
          {
            foafEntity = authorFactory.create(new QName(uri));
            map.put(uri, foafEntity);
//            foafEntity.getLookup().lookup(Concept.class).setRdfsLabel(name);
            foafEntity.getLookup().lookup(org.openrdf.concepts.foaf.Person.class).getFoafGivennames().add(name);
          }

        return foafEntity;
      }
  }
